import os
import re
import logging
import platform
from collections.abc import Callable
from dataclasses import dataclass
from functools import lru_cache, wraps
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal
from urllib.parse import parse_qs, urljoin, urlparse, urlsplit

from lxml import html

from .constant import BASE_URL, VALID_PAGE

if TYPE_CHECKING:
    pass

logger = logging.getLogger()
VALID_CLIENT = Literal["httpx", "curl", "primp"]


@dataclass
class BaseConfig:
    album_url: str
    album_dir: Path
    start_page: int = 1
    download: bool = True


class AlbumTracker:
    """Download log in units of albums."""

    def __init__(self, download_log: str | Path) -> None:
        self.album_log_path = Path(download_log)

    def is_downloaded(self, album_url: str) -> bool:
        if os.path.exists(self.album_log_path):
            with open(self.album_log_path) as f:
                downloaded_albums = f.read().splitlines()
            return album_url in downloaded_albums
        return False

    def log_downloaded(self, album_url: str) -> None:
        if not self.is_downloaded(album_url):
            with open(self.album_log_path, "a") as f:
                f.write(album_url + "\n")


def parse_album_urls(html_content: str) -> list[tuple[str, str]]:
    tree = html.fromstring(html_content)
    album_links = tree.xpath(
        '//div[contains(@class, "row gutter-10 albums-list")]//a[contains(@class, "media-cover")]'
    )
    result = []
    for link in album_links:
        album_url = urljoin(BASE_URL, link.get("href"))
        img_tag = link.xpath('.//img[contains(@class, "card-img-top")]')
        album_alt = img_tag[0].get("alt", "").strip() if img_tag else "Untitled"
        result.append((album_url, album_alt))
    return result


def parse_photo_urls(html_content: str) -> list[Any]:
    path = '//div[@class="photos-list text-center"]//div[@class="album-photo my-2"]//img/@data-src'
    tree = html.fromstring(html_content)
    return tree.xpath(path)


def parse_album_title(html_content: str) -> str:
    path = '//div[contains(@class, "card-body")]//h1[contains(@class, "h5 text-center mb-3")]'
    tree = html.fromstring(html_content)
    title_element = tree.xpath(path)
    return title_element[0].text_content().strip() if title_element else ""


def parse_next_page_url(html_content: str) -> None | str:
    path_current_page = '//ul[contains(@class, "pagination")]//li[contains(@class, "active")]'
    path_next_page_item = 'following-sibling::li[contains(@class, "page-item")][1]/a'
    tree = html.fromstring(html_content)
    current_page = tree.xpath(path_current_page)
    if current_page:
        next_page_item = current_page[0].xpath(path_next_page_item)
        if next_page_item:
            return urljoin(BASE_URL, next_page_item[0].get("href"))
    return None


def parse_url_mode(url: str, valid_pages: tuple[str, ...] = VALID_PAGE) -> str:
    if not url.startswith(BASE_URL):
        raise ValueError(f"URL must start with {BASE_URL}, got {url}")

    parsed_url = urlsplit(url)
    path_segments = parsed_url.path.strip("/").split("/")

    if not path_segments:
        raise ValueError(f"Invalid URL: {url}")

    mode = path_segments[0]
    if mode not in valid_pages:
        raise ValueError(f"Unsupported mode: {mode}")

    return mode


def parse_page_num(url: str) -> int:
    """parse page url, default is 1"""
    parsed_url = urlsplit(url)
    query_params = parse_qs(parsed_url.query)
    return int(query_params.get("page", ["1"])[0])


def parse_input_url(input_path: str) -> list[str]:
    """parse argparse download inputs, accept multiple urls input or multiple txt input"""

    # Check if input is a valid url, wrap it into a list
    if is_valid_url(input_path):
        return [input_path]

    # If input not a url, check if the file path exists.
    # Validate each url if the file exists, otherwise, return an empty list.
    try:
        with open(input_path) as f:
            return [line.strip() for line in f if is_valid_url(line.strip())]
    except Exception:
        return []


def is_valid_url(url: str) -> bool:
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except Exception:
        return False


def get_system_config_dir() -> Path:
    """Return the config directory."""
    if platform.system() == "Windows":
        base = os.getenv("APPDATA", "")
    else:
        base = os.path.expanduser("~/.config")
    return Path(base) / "v2dl"


def find_cookie_files(config_dir: str | Path) -> list[str]:
    """Find all cookie files with 'cookie' in their names."""
    config_dir = Path(config_dir)
    return [
        str(file)
        for file in config_dir.iterdir()
        if file.is_file() and "cookie" in file.name and file.suffix == ".txt"
    ]


def suppress_log(log_level: int) -> None:
    level = logging.DEBUG if log_level == logging.DEBUG else logging.WARNING
    logging.getLogger("httpx").setLevel(level)
    logging.getLogger("httpcore").setLevel(level)


@lru_cache
def mkdir(path: Path) -> None:
    path.mkdir(parents=True, exist_ok=True)


def add_underscore_before_first_number(input_string: str) -> str:
    return re.sub(r"(?=\d)", "_", input_string, count=1)


def auto_await(func: Callable[..., Any]) -> Callable[..., Any]:
    # TODO: blocking and run async function or run normal function
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        result = func(*args, **kwargs)
        raise NotImplementedError(str(result) + "auto await wrapper not yet implemented")

    return wrapper



BASE_URL = "https://www.v2ph.com"
BASE_DIR = "v2dl"
SLEEP_TIME = 5
SPEED_LIMIT_KBPS = 400
MAX_WORKER = 5
DOWNLOAD_LOG = "downloaded_albums.txt"

BROWSER = "chrome124"  # for primp/curl-cffi
UA = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.6367.207 Safari/537.36"
HEADERS = {
    "Referer": BASE_URL,
    "User-Agent": UA,
    "Accept-Language": "zh-TW",
}
VALID_PAGE = ("album", "actor", "company", "category", "country")



import sys
import logging
import argparse
from pathlib import Path

from .constant import BASE_DIR
from .utils import get_system_config_dir
from .version import __version__


class CustomHelpFormatter(argparse.RawTextHelpFormatter):
    def __init__(self, prog) -> None:  # type: ignore
        super().__init__(prog, max_help_position=36)

    def _format_action_invocation(self, action):  # type: ignore
        if not action.option_strings:
            (metavar,) = self._metavar_formatter(action, action.dest)(1)
            return metavar
        else:
            parts = []
            # if the Optional doesn't take a value, format is:
            #    -s, --long
            if action.nargs == 0:
                parts.extend(action.option_strings)

            # if the Optional takes a value, format is:
            #    -s ARGS, --long ARGS
            # change to
            #    -s, --long ARGS
            else:
                default = action.dest.upper()
                args_string = self._format_args(action, default)
                for option_string in action.option_strings:
                    # parts.append('%s %s' % (option_string, args_string))
                    parts.append(f"{option_string}")
                parts[-1] += f" {args_string}"
            return ", ".join(parts)


def parse_argument() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Lite version of V2PH-Downloader",
        formatter_class=CustomHelpFormatter,
    )

    parser.add_argument(
        "inputs",
        nargs="*",
        help="URLs or text files containing URLs (one per line)",
    )

    parser.add_argument(
        "--force",
        "-f",
        action="store_true",
        help="Force downloading, not skipping downloaded albums",
    )

    parser.add_argument(
        "--download-dir",
        "-d",
        default=str(Path.home() / "Downloads" / BASE_DIR),
        help=f"Download directory (default: ~/Downloads/{BASE_DIR})",
    )

    parser.add_argument(
        "--cookies-path",
        "-c",
        default=get_system_config_dir(),
        help=f"Path to cookies directory (default: ~/.config/{BASE_DIR})",
    )

    parser.add_argument(
        "--language",
        "-l",
        default="zh-TW",
        help="Preferred language, used for naming the download directory (default: zh-TW)",
    )

    parser.add_argument(
        "--session",
        "-s",
        default="curl",
        choices=["curl", "httpx"],
        help="The HTTP client session to use (default: curl)",
    )

    parser.add_argument(
        "--log-level",
        default=2,
        type=int,
        choices=range(1, 6),
        help="Set log level from low to high (1~5)",
    )

    parser.add_argument(
        "--version",
        "-v",
        action="store_true",
        help="Show package version",
    )

    args = parser.parse_args()

    if args.version:
        print(__version__)  # noqa: T201
        sys.exit(0)

    if not args.inputs:
        parser.error("the following arguments are required: inputs")

    if args.log_level is not None:
        log_level_mapping = {
            1: logging.DEBUG,
            2: logging.INFO,
            3: logging.WARNING,
            4: logging.WARNING,
            5: logging.CRITICAL,
        }
        args.log_level = log_level_mapping.get(args.log_level, logging.INFO)

    return args



import os
import time
import asyncio
import logging
from abc import ABC, abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from http.cookiejar import MozillaCookieJar
from pathlib import Path
from typing import TYPE_CHECKING, Any
from urllib.parse import urljoin

from httpx import AsyncClient, HTTPError

# from requests.cookies import cookiejar_from_dict
from .constant import BASE_URL, BROWSER, HEADERS, SPEED_LIMIT_KBPS
from .utils import add_underscore_before_first_number, auto_await, mkdir

if TYPE_CHECKING:
    import httpx
    import curl_cffi.requests as curl_

    from ._types import Dl_Status, Valid_Session


logger = logging.getLogger()


@dataclass
class DownloadConfig:
    download_dir: Path
    semaphore: asyncio.Semaphore
    session: "BaseSession"
    speed_limit_kbps: int = SPEED_LIMIT_KBPS
    start_idx: int = 1
    force_download: bool = False


class BaseResponse(ABC):
    @abstractmethod
    def json(self) -> Any:
        pass

    @property
    @abstractmethod
    def text(self) -> str:
        pass

    @abstractmethod
    def raise_for_status(self) -> None:
        pass

    @property
    @abstractmethod
    def url(self) -> Any:
        pass


class HttpxResponse(BaseResponse):
    def __init__(self, response: "httpx.Response") -> None:
        self._response = response

    @property
    def text(self) -> str:
        return self._response.text

    @property
    def url(self) -> "httpx.URL":
        return self._response.url

    def json(self) -> Any:
        return self._response.json()

    def raise_for_status(self) -> None:
        _ = self._response.raise_for_status()


class CurlResponse(BaseResponse):
    def __init__(self, response: "curl_.Response") -> None:
        self._response = response

    @property
    def text(self) -> str:
        return self._response.text

    @property
    def url(self) -> str:
        return self._response.url

    def json(self) -> Any:
        return self._response.json()

    def raise_for_status(self) -> None:
        self._response.raise_for_status()


class PrimpResponse(BaseResponse):
    def __init__(self, response: "curl_.Response") -> None:
        self._response = response

    @property
    def text(self) -> str:
        return self._response.text

    @property
    def url(self) -> str:
        return self._response.url

    def json(self) -> Any:
        return self._response.json()

    def raise_for_status(self) -> None:
        if self._response.status_code >= 400:
            raise HTTPError(f"HTTP Error (primp): {self._response.status_code}")


class BaseSession(ABC):
    def __init__(
        self,
        cookies_path: str,
        headers: dict[str, str],
        timeout: int = 15,
    ) -> None:
        self.cookies, self.headers = prepare_session(cookies_path, headers)
        self.timeout = timeout

    @abstractmethod
    async def get(self, url: str, *args: Any, **kwargs: Any) -> BaseResponse: ...

    @auto_await
    @abstractmethod
    async def sync_get(self, url: str, *args: Any, **kwargs: Any) -> BaseResponse:
        raise NotImplementedError("auto_await not implemented yet")

    @abstractmethod
    async def post(
        self, url: str, data: dict[Any, Any], *args: Any, **kwargs: Any
    ) -> BaseResponse: ...

    @abstractmethod
    async def stream(
        self,
        semaphore: asyncio.Semaphore,
        url: str,
        dest: Path,
        speed_limit_kbps: int,
        *args: Any,
        **kwargs: Any,
    ) -> "Dl_Status":
        """Download with speed limit using an existing session.

        Args:
            semaphore (asyncio.Semaphore): The semaphore to restrict the concurrency
            client (Client): The request client session, can be a httpx/curl-cffi/primp session
            url (str): The download url
            dest (Path): The download destination
            headers (dict[str, str]): The header for requests client
            speed_limit_kbps (int): The download speed limit
        Returns:
            dl_status (tuple[bool, str, str]): The download status, a tuple with first term indicates
            download status, the second term is the download url and the last term is the destination.
        """


class HttpxSession(BaseSession):
    def __init__(
        self,
        cookies_path: str,
        headers: dict[str, str],
        timeout: int,
    ) -> None:
        super().__init__(cookies_path, headers, timeout)
        self.session = AsyncClient(headers=headers, cookies=self.cookies, http2=True, timeout=15)

    async def get(self, url: str, *args: Any, **kwargs: Any) -> BaseResponse:
        response = await self.session.get(url, *args, **kwargs)
        return HttpxResponse(response)

    @auto_await
    async def sync_get(self, url: str, *args: Any, **kwargs: Any) -> BaseResponse:
        response = await self.session.get(url, *args, **kwargs)
        return HttpxResponse(response)

    async def post(self, url: str, data: dict[Any, Any], *args: Any, **kwargs: Any) -> BaseResponse:
        response = await self.session.post(url, *args, json=data, **kwargs)
        return HttpxResponse(response)

    async def stream(
        self,
        semaphore: asyncio.Semaphore,
        url: str,
        dest: Path,
        speed_limit_kbps: int,
        *args: Any,
        **kwargs: Any,
    ) -> "Dl_Status":
        chunk_size = kwargs.get("chunk_size", 0)
        if speed_limit_kbps <= 0:
            raise ValueError("Speed limit must be a positive number.")

        speed_limit_bps = speed_limit_kbps * 1024

        async with semaphore:
            async with self.session.stream("GET", url, timeout=30.0) as response:
                response.raise_for_status()
                with open(dest, "wb") as file:
                    downloaded = 0
                    start_time = time.time()

                    async for chunk in response.aiter_bytes(chunk_size=chunk_size):
                        if not chunk:
                            break
                        file.write(chunk)
                        downloaded += len(chunk)
                        elapsed_time = time.time() - start_time
                        expected_time = downloaded / speed_limit_bps
                        if elapsed_time < expected_time:
                            await asyncio.sleep(expected_time - elapsed_time)

        return download_check(url, dest, downloaded)


class CurlSession(BaseSession):
    def __init__(
        self,
        cookies_path: str,
        headers: dict[str, str],
        timeout: int,
    ) -> None:
        super().__init__(cookies_path, headers, timeout)
        try:
            from curl_cffi.requests import AsyncSession

            self.session = AsyncSession(
                cookies=self.cookies, headers=self.headers, timeout=15, impersonate=BROWSER
            )
        except ModuleNotFoundError:
            raise ImportError("curl_requests module is required for CurlCffiAdapter.")

    async def get(self, url: str, *args: Any, **kwargs: Any) -> BaseResponse:
        response = await self.session.get(url, **kwargs)
        return CurlResponse(response)

    @auto_await
    async def sync_get(self, url: str, *args: Any, **kwargs: Any) -> BaseResponse:
        response = await self.session.get(url, *args, **kwargs)
        return CurlResponse(response)

    async def post(self, url: str, data: dict[Any, Any], *args: Any, **kwargs: Any) -> BaseResponse:
        response = await self.session.post(url, json=data, **kwargs)
        return CurlResponse(response)

    async def stream(
        self,
        semaphore: asyncio.Semaphore,
        url: str,
        dest: Path,
        speed_limit_kbps: int,
        *args: Any,
        **kwargs: Any,
    ) -> "Dl_Status":
        async with semaphore:
            if speed_limit_kbps <= 0:
                raise ValueError("Speed limit must be a positive number.")

            speed_limit_bps = speed_limit_kbps * 1024

            response = await self.session.get(url, stream=True)
            response.raise_for_status()

            with open(dest, "wb") as f:
                downloaded = 0
                start_time = time.time()

                async for chunk in response.aiter_content():  # chunk size not possible for curl
                    if not chunk:
                        break
                    f.write(chunk)
                    downloaded += len(chunk)

                    elapsed_time = time.time() - start_time
                    expected_time = downloaded / speed_limit_bps
                    if elapsed_time < expected_time:
                        await asyncio.sleep(expected_time - elapsed_time)

        return download_check(url, dest, downloaded)


# TODO: Wait for AsyncClient feature https://github.com/deedy5/primp
class PrimpSession(BaseSession):
    def __init__(
        self,
        cookies_path: str,
        headers: dict[str, str],
        timeout: int,
    ) -> None:
        super().__init__(cookies_path, headers, timeout)
        try:
            import primp  # type: ignore

            browser = add_underscore_before_first_number(BROWSER)
            # headers will be ignored with impersonate is set
            self.session = primp.Client(cookies=self.cookies, impersonate=browser, http2=True)  # type: ignore
        except ModuleNotFoundError:
            raise ImportError(
                "primp module not found, please install primp with `pip install v2dl-lite[ja4]`."
            )

    async def get(self, url: str, *args: Any, **kwargs: Any) -> BaseResponse:
        raise NotImplementedError(
            "primp method is under developing, see https://github.com/deedy5/primp"
        )

    @auto_await
    def sync_get(self, url: str, *args: Any, **kwargs: Any) -> BaseResponse:
        response = self.session.get(url, *args, **kwargs)
        return PrimpResponse(response)

    async def post(self, url: str, data: dict[Any, Any], *args: Any, **kwargs: Any) -> BaseResponse:
        raise NotImplementedError(
            "primp method is under developing, see https://github.com/deedy5/primp"
        )

    async def stream(
        self,
        semaphore: asyncio.Semaphore,
        url: str,
        dest: Path,
        speed_limit_kbps: int,
        chunk_size: int,
        *args: Any,
        **kwargs: Any,
    ) -> "Dl_Status":
        raise NotImplementedError


def create_session(
    client: "Valid_Session",
    cookies_path: str,
    headers: dict[str, str] = HEADERS,
    timeout: int = 15,
    **kwargs: dict[Any, Any],
) -> BaseSession:
    client_map: dict[Valid_Session, Callable[..., Any]] = {
        "httpx": lambda: HttpxSession(cookies_path, headers, timeout, **kwargs),
        "curl": lambda: CurlSession(cookies_path, headers, timeout, **kwargs),
        "primp": lambda: PrimpSession(cookies_path, headers, timeout, **kwargs),
    }

    try:
        return client_map[client]()
    except KeyError:
        raise ValueError(f"Unknown session_type: {client}")


async def download_photos(
    config: "DownloadConfig",
    photo_urls: list[str],
) -> list["Dl_Status"]:
    """An async job submitter for download function"""

    download_tasks = []
    idx = config.start_idx

    for photo_url in photo_urls:
        file_name = f"{idx:03d}.{photo_url.split('.')[-1]}"
        dest = config.download_dir / file_name

        if not config.force_download:
            if dest.is_file():
                continue

        mkdir(config.download_dir)
        task = config.session.stream(
            semaphore=config.semaphore,
            client=config.session,
            url=photo_url,
            dest=dest,
            speed_limit_kbps=config.speed_limit_kbps,
            chunk_size=4096,
        )
        download_tasks.append(task)
        idx += 1

    results = await asyncio.gather(*download_tasks)
    for r in results:
        if r[0]:
            logger.debug(f"Download success: {r[1]}")
        else:
            logger.error(f"Download fail, url: {r[1]}, dest: {r[2]}")
    return results


def download_check(url: str, dest: str | Path, downloaded: float) -> "Dl_Status":
    if os.path.exists(dest):
        actual_size = os.path.getsize(dest)
        lower_bound = downloaded * 0.9999
        upper_bound = downloaded * 1.0001
        if lower_bound <= actual_size <= upper_bound:
            return (True, url, str(dest))
    return (False, url, str(dest))


def prepare_session(cookie_file: str, headers: dict[str, str]) -> tuple[Any, dict[str, str]]:
    if os.path.isfile(cookie_file):
        cookie_jar = MozillaCookieJar(cookie_file)
        cookie_jar.load()
        cookies_dict = {c.name: c.value for c in cookie_jar if c.value is not None}
    else:
        logger.warning(f"cookie file not found: {cookie_file}")
        cookies_dict = {}
    # cookies = cookiejar_from_dict(cookies_dict)
    return cookies_dict, headers


def access_fail(response: BaseResponse, msg: str | None = None) -> None:
    try:
        response.raise_for_status()
    except HTTPError as e:
        # Only catch response error, other exceptions will be passed to upper function
        raise HTTPError(f"HTTP request failed: {e} {msg or ''}")

    if login_fail(response):
        raise LoginRequiredError("LoginRequiredError: Login failed, please update your cookies")


def login_fail(response: BaseResponse) -> bool:
    return response.url == urljoin(BASE_URL, "login")


class LoginRequiredError(Exception):
    """Check login error"""


class ConnectError(Exception):
    """Network status code error"""



import time
import random
import asyncio
from logging import getLogger
from pathlib import Path
from typing import TYPE_CHECKING

from .constant import BASE_URL, DOWNLOAD_LOG, SLEEP_TIME
from .downloader import (
    DownloadConfig,
    LoginRequiredError,
    access_fail,
    create_session,
    download_photos,
)
from .utils import (
    AlbumTracker,
    BaseConfig,
    get_system_config_dir,
    parse_album_title,
    parse_album_urls,
    parse_input_url,
    parse_next_page_url,
    parse_page_num,
    parse_photo_urls,
    parse_url_mode,
)

if TYPE_CHECKING:
    from ._types import Dl_Status, Valid_Session


logger = getLogger()


class Scrapper:
    """Asynchronous downloader for V2PH.com

    A scraper that handles downloading albums and photos from V2PH.com
    with support for multiple cookie files, concurrent downloads, and
    tracking of previously downloaded albums.

    Attributes:
        inputs (list[str]): List of input URLs or text files to download from
        download_dir (Path): Directory to save downloaded albums
        force_download (bool): Flag to force re-downloading of existing albums
        headers (dict[str, str]): HTTP headers for requests
        max_worker (int): Maximum number of concurrent download workers
        album_tracker (AlbumTracker): Tracks downloaded albums to prevent duplicates
    """

    def __init__(
        self,
        inputs: list[str] | str,
        cookie_files: list[str] | str,
        download_dir: str,
        force_download: bool,
        session_type: "Valid_Session",
        headers: dict[str, str],
        max_worker: int,
    ) -> None:
        """Initialize Scrapper

        Args:
            inputs (list[str]): Download target from user input. Can be a list of txt file or urls.
            cookie_files (list[str]): Cookie files extracted from `find_cookie_files`.
            force_download (bool): Whether to skip downloaded url.
            headers: The header used for cloudscraper and httpx.
            max_worker: The maximum concurrency of the async download function.
        """
        self.inputs = inputs
        self.download_dir = Path(download_dir)
        self.force_download = force_download
        self.session_type: Valid_Session = session_type  # suppress IDE
        self.headers = headers
        self.max_worker = max_worker

        log_dir = get_system_config_dir() / DOWNLOAD_LOG
        self.album_tracker = AlbumTracker(log_dir)

        if not isinstance(self.inputs, list):
            self.inputs = [self.inputs]
        if not isinstance(cookie_files, list):
            self.cookie_files = [cookie_files]
        else:
            self.cookie_files = cookie_files

    async def run(self) -> None:
        """extract user input, looping around all the txt/urls and calls scrape method

        Iterates through inputs, skips already downloaded albums unless force_download is True,
        and triggers album scraping for each valid URL.
        """
        for input_ in self.inputs:
            urls = parse_input_url(input_)
            if not urls:
                logger.error(f"Warning: No valid URLs in {input_}")
                continue

            for url in urls:
                if self.album_tracker.is_downloaded(url) and not self.force_download:
                    logger.info(f"Album {url} already downloaded, skipping.")
                    continue
                await self.scrape(url)

    async def scrape(self, url: str) -> list["Dl_Status"]:
        """Input check and error management for a single url with multiple cookie files.

        Args:
            url (str): The processing url
        Return:
            Dl_Status: The download status, including a successful flag, url and destination
        """
        cookie_fail = False
        mode = parse_url_mode(url)

        for cookie_file in self.cookie_files:
            # httpx can't pass cloudflare challenge
            if self.session_type == "httpx":
                self.cf_sess = create_session("curl", cookie_file, self.headers)
                self.httpx_sess = create_session(self.session_type, cookie_file, self.headers)

            # primp doesn't support async
            elif self.session_type == "primp":
                self.cf_sess = create_session(self.session_type, cookie_file, self.headers)
                self.httpx_sess = create_session("curl", cookie_file, self.headers)

            # curl-cffi support both
            else:
                self.cf_sess = create_session(self.session_type, cookie_file, self.headers)
                self.httpx_sess = self.cf_sess

            if cookie_fail:
                logger.info(f"Login fail. Switching to another cookie file {cookie_file}")
            else:
                logger.debug(f"Downloading with cookie file {Path(cookie_file).name}")

            try:
                await self.__warmup()
                # self.__transfer_parameters()
                return await self.scrape_selector(url=url, mode=mode)
            except LoginRequiredError as e:
                logger.error(e)
                cookie_fail = True
            except Exception as e:
                raise RuntimeError(f"RuntimeError: {e}") from e
        return []

    async def scrape_selector(self, url: str, mode: str) -> list["Dl_Status"]:
        """Select the scrape mode"""
        results = []

        if mode == "album":
            response = await self.cf_sess.get(url)
            access_fail(response)
            start_page = parse_page_num(url)

            album_alt = parse_album_title(response.text)
            album_dir = self.download_dir / album_alt
            logger.info(f"Start scraping photos from {url}, album title: {album_alt}")

            base_config = BaseConfig(
                album_url=url,
                album_dir=album_dir,
                start_page=start_page,
                download=True,
            )
            results = await self.scrape_photo_urls(base_config)
        else:
            album_urls_with_alts = await self.scrape_album_urls(url)
            start_page = 1
            for album_url, album_alt in album_urls_with_alts:
                album_dir = self.download_dir / album_alt

                base_config = BaseConfig(
                    album_url=album_url,
                    album_dir=album_dir,
                    start_page=start_page,
                    download=True,
                )
                results.extend(await self.scrape_photo_urls(base_config))

        return results

    async def scrape_album_urls(self, start_url: str) -> list[tuple[str, str]]:
        """The mode that scrape the url of albums"""

        current_url: str | None = start_url
        album_urls_with_alts = []

        while current_url:
            response = await self.cf_sess.get(current_url)
            access_fail(response)

            logger.info(f"Start scraping urls from {start_url}")
            album_urls_with_alts.extend(parse_album_urls(response.text))
            current_url = parse_next_page_url(response.text)

            sleep_time = random.uniform(SLEEP_TIME, SLEEP_TIME * 2)
            logger.info(f"Sleep {sleep_time:.1f}s to avoid bot detection")
            time.sleep(sleep_time)

        return album_urls_with_alts

    async def scrape_photo_urls(self, config: BaseConfig) -> list["Dl_Status"]:
        """The mode that scrape the url of photos"""

        idx = 10 * (config.start_page - 1) + 1  # 10 images per page
        current_url: str | None = config.album_url
        semaphore = asyncio.Semaphore(self.max_worker)
        results = []

        while current_url:
            response = await self.cf_sess.get(current_url)
            access_fail(response)
            # self.__transfer_parameters()

            urls = parse_photo_urls(response.text)

            if config.download:
                download_config = DownloadConfig(
                    download_dir=config.album_dir,
                    semaphore=semaphore,
                    session=self.httpx_sess,
                    start_idx=idx,
                    force_download=self.force_download,
                )
                results.extend(await download_photos(download_config, urls))
                idx += len(urls)

            current_url = parse_next_page_url(response.text)

            sleep_time = random.uniform(SLEEP_TIME, SLEEP_TIME * 4)
            logger.info(f"Sleep {sleep_time:.1f}s to avoid bot detection")
            time.sleep(sleep_time)

        self.album_tracker.log_downloaded(config.album_url)
        return results

    async def __warmup(self) -> None:
        warmup_times = random.randint(1, 1)
        warmup_url = [
            "https://www.v2ph.com/actor/Umi-Shinonome",
            "https://www.v2ph.com/country/japan",
            "https://www.v2ph.com/category/magazine",
            "https://www.v2ph.com/category/short-hair",
        ]

        logger.info("warming up session")
        for _ in range(warmup_times):
            r = await self.cf_sess.get(random.choice(warmup_url) + f"?page={random.randint(1, 20)}")
            access_fail(r, "Error occurs at: warmup")
            time.sleep(random.uniform(0.1, 3.0))

    def __transfer_parameters(self) -> None:
        self.httpx_sess.headers.update(dict(self.cf_sess.headers))  # type: ignore
        self.httpx_sess.cookies.clear()
        for k, v in dict(self.cf_sess.cookies).items():
            self.httpx_sess.cookies.set(k, v, BASE_URL)



import sys
import asyncio
from pathlib import Path

from v2dl_lite.constant import HEADERS, MAX_WORKER
from v2dl_lite.logger import setup_logging
from v2dl_lite.options import parse_argument
from v2dl_lite.scrapper import Scrapper
from v2dl_lite.utils import find_cookie_files, suppress_log


def main() -> int:
    headers = HEADERS
    max_worker = MAX_WORKER
    args = parse_argument()
    cookie_files = find_cookie_files(Path(args.cookies_path))
    logger = setup_logging(level=args.log_level)
    suppress_log(args.log_level)

    if not cookie_files:
        return 1

    if args.language:
        headers["Accept-Language"] = args.language

    # chr(10): "\n"
    logger.info(f"""
Your download configuration:
----------------------------------------------------------
Download base folder: {args.download_dir}
{chr(10).join(f"Cookies file: {file}" for file in cookie_files)}
Skipping downloaded albums? {"Yes" if not args.force else "No"}
----------------------------------------------------------
""")

    if sys.platform == "win32":
        try:
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
        except Exception as e:
            logger.warning(f"Curl-cffi initialization error: {e}")

    scrapper = Scrapper(
        args.inputs,
        cookie_files,
        args.download_dir,
        args.force,
        args.session,
        headers,
        max_worker,
    )
    asyncio.run(scrapper.run())

    return 0


if __name__ == "__main__":
    raise SystemExit(main())
